@@ -43,6 +43,7 @@
 #include "dqnet_priv.h"
 #include <net/netfilter/nf_conntrack_expect.h>
 #include <net/netfilter/nf_conntrack_helper.h>
+#include <net/netfilter/nf_conntrack_acct.h>
 #include "mso_feat.h"
 
 DEFINE_SPINLOCK(flowmgr_map_lock);
@@ -827,7 +828,7 @@ int inner_map_flush(int domain)
 		memset(&params, 0, sizeof(params));
 		params.domain_index = domain;
 
-		ret = flowmgr_config_mapt(&params);
+		ret = flowmgr_config_mapt(&params, 0);
 		if (ret != 0)
 			FLOWMGR_MAP_RETURN(-1)
 
@@ -839,6 +840,7 @@ int inner_map_flush(int domain)
 	if (db && (db->state == 1)) {
 		// Set state to unconfigure
 		db->state = 0;
+		db->id = -1;
 	}
 
 	return 0;
@@ -1105,7 +1107,7 @@ int hw_config_map(struct sk_buff *skb,
 
 		// Configure
 		if (flowmgr_is_feat_map_enable() && flowmgr.disable_mapt_accel < 3)
-			ret = flowmgr_config_mapt(&(db->mapt));
+			ret = flowmgr_config_mapt(&(db->mapt), dev->ifindex);
 
 		if (ret != 0)
 			FLOWMGR_MAP_RETURN(-1)
@@ -1113,6 +1115,7 @@ int hw_config_map(struct sk_buff *skb,
 		// Set state to configured
 		db->state = 1;
 		db->oif = dev->ifindex;
+		db->id = domain;
 
 		flowmgr_dump_mapt(&(db->mapt));
 
@@ -1376,7 +1379,7 @@ int inner_promote_mapt(int prot,
 
 	if (db->state == 0) {
 		if (ct_offload_info->debug)
-			pr_alert("FLOWMGR: Promotion error - MAPT tunnel not intialized in hw\n");
+			pr_alert("FLOWMGR: Promotion error - MAPT tunnel not initialized in hw\n");
 		return 0;
 	}
 
@@ -1713,3 +1716,70 @@ bool flowmgr_is_feat_map_enable(void)
 	}
 	return enable;
 }
+
+int flowmgr_update_stats_mapt(const struct nf_conn *ct,
+			      int direction,
+			      unsigned int packets_diff,
+			      unsigned int bytes_diff)
+{
+	struct nf_conn_offload *ct_offload;
+	struct nf_conn *master_ct = NULL;
+	struct flowmgr_map_offload *map_offload = NULL;
+	struct nf_conn_acct *acct;
+	int domain;
+	struct flowmgr_tdb_entry *tdb;
+	struct net_device *map_dev;
+
+	if (!ct || nf_ct_is_dying(ct))
+		return 0;
+
+	ct_offload = nf_conn_offload_find(ct);
+
+	if (ct_offload_orig.map)
+		map_offload = (struct flowmgr_map_offload *) ct_offload_orig.map;
+	else if (ct_offload_repl.map)
+		map_offload = (struct flowmgr_map_offload *) ct_offload_repl.map;
+	else
+		return 0;
+
+	master_ct = map_offload->master_ct;
+	if (!master_ct || nf_ct_is_dying(master_ct))
+		return 0;
+
+	if (map_offload->params.type == ft_ipv6) {
+		domain = map_offload->params.ipv6.mapt.domain_index;
+		tdb = flowmgr_tunnel_find_mapt_domain(domain);
+		if (tdb) {
+			map_dev = __dev_get_by_index(&init_net, tdb->ifi);
+			if (map_dev) {
+				/* nat46 driver stats */
+				map_dev->stats.rx_packets += packets_diff;
+				map_dev->stats.rx_bytes += bytes_diff;
+			}
+		}
+	} else {
+		domain = map_offload->params.ipv4.mapt.domain_index;
+		tdb = flowmgr_tunnel_find_mapt_domain(domain);
+		if (tdb) {
+			map_dev = __dev_get_by_index(&init_net, tdb->ifi);
+			if (map_dev) {
+				/* nat46 driver stats */
+				map_dev->stats.tx_packets += packets_diff;
+				map_dev->stats.tx_bytes += bytes_diff;
+			}
+		}
+	}
+
+	/* Update nf connection acc stats for master ct*/
+	acct = nf_conn_acct_find(master_ct);
+	if (acct) {
+		struct nf_conn_counter *nf_counter;
+		nf_counter = &(acct->counter[direction]);
+		atomic64_add(packets_diff,
+			     &nf_counter->packets);
+		atomic64_add(bytes_diff,
+			     &nf_counter->bytes);
+	}
+	flowmgr_update_offload_retry_timeout(master_ct);
+	return 0;
+}
